#include "stdafx.h"
#include "resource.h"

#include "str.h"

#include "MySock.h"
#include "MyConf.h"

#include "cehttp.h"
#include "httpd.h"
#include "httphelper.h"

#if !defined(_WIN32_WCE)
#include "shellapi.h"
//#include "malloc.h"		// for free()
#endif

HINSTANCE	g_hInst;
extern MyConf	g_conf;

static BOOL ShowSetting(HWND hDlg);
static BOOL SetSetting(HWND hDlg);
static BOOL CALLBACK DlgProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK SettingDlgProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK HelpDlgProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static BOOL MakeLand(HWND hDlg);


//-----------------------------------------------------------------------------
//	@detail
//		WinMain function.
//	@param
//	@return
//-----------------------------------------------------------------------------
#if !defined(_WIN32_WCE)
int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow )
#else
int WINAPI WinMain(	HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPTSTR    lpCmdLine,
					int       nCmdShow)
#endif
{
	g_hInst = hInstance;
    DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAINDLG), NULL, DlgProc);
    return 0;

}

//-----------------------------------------------------------------------------
//	@detail
//		Main dialog box
//	@param
//	@return
//-----------------------------------------------------------------------------
static BOOL CALLBACK DlgProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	static	LISTEN_THREAD* pListenThreadPack = NULL;
	static	NOTIFYICONDATA	nIcon;
	static	HICON			hIconTaskBar;
	static int nShutdownGoing = 0;


	switch(Msg)
	{
	case WM_INITDIALOG:
	{
		MySock s;
		pListenThreadPack = NULL;
		char szTmpString[256] = {0};

		if( !MakeLand(hDlg) )
		{
			DialogBox(g_hInst, MAKEINTRESOURCE(IDD_DOCSETTING), NULL, SettingDlgProc);
		}

		// ------------------------------------------------------
		// Print Version no.
		// ------------------------------------------------------
		WCHAR* pw = malloc_char2wchar( MYCONF_STR_SRVNAME );
		if( pw != NULL )
		{
			SetWindowText( hDlg, pw );
			FREEIF( pw );
		}
		
		// ------------------------------------------------------
		// Print IP address.
		// ------------------------------------------------------
		

		int nResult = gethostname( szTmpString, 255 );

		if( nResult != 0 ||  szTmpString[0] == '\0' )
		{
			SetDlgItemText(hDlg, IDC_IPADDR, TEXT("unknown."));
		}
		else
		{
			HOSTENT* lphostent = gethostbyname( szTmpString );
			if( lphostent == NULL )
			{
				SetDlgItemText(hDlg, IDC_IPADDR, TEXT("unknown"));
			}
			else
			{
				WCHAR szIpAddr[16];
				char* adr = *(lphostent->h_addr_list);
				wsprintf(szIpAddr, TEXT("%d.%d.%d.%d"), (UCHAR)adr[0], (UCHAR)adr[1], (UCHAR)adr[2], (UCHAR)adr[3] );
				SetDlgItemText(hDlg, IDC_IPADDR, szIpAddr);
			}
		}

		// ------------------------------------------------------
		// Stating server threads
		// ------------------------------------------------------

		if( NULL == (pListenThreadPack = (LISTEN_THREAD*)MALLOC( sizeof(LISTEN_THREAD)) ) ) return FALSE;
		pListenThreadPack->bValid = FALSE;			// cold start
		pListenThreadPack->hWnd = hDlg;				// hWnd for message

		DWORD ThreadID;
		pListenThreadPack->hThread = CreateThread(NULL, 0, 
			(LPTHREAD_START_ROUTINE)thListen, (LPVOID)pListenThreadPack, 0, &ThreadID);

		// ------------------------------------------------------
		// Make Tasktray icon
		// ------------------------------------------------------

		nIcon.cbSize			= sizeof(nIcon);
		nIcon.hIcon				= (HICON)LoadImage(
										g_hInst,
										MAKEINTRESOURCE(IDI_ICON_MAIN),
										IMAGE_ICON, 16,16,0);
		nIcon.hWnd				= hDlg;
		nIcon.uCallbackMessage	= MY_NOTIFYICON;
		nIcon.uFlags			= NIF_MESSAGE | NIF_ICON;
		nIcon.uID				= 1;

		Shell_NotifyIcon(NIM_ADD, &nIcon);

		ShowWindow(hDlg, SW_HIDE);

		return TRUE;
	}

	// ----------------------------------------------------
	// Task tray icon notify
	case MY_NOTIFYICON:
		switch(lParam)
		{
		case WM_LBUTTONDOWN:
		{
			if(IsWindowVisible(hDlg) )
			{
				ShowWindow(hDlg, SW_HIDE);
			}
			else
			{
				ShowWindow(hDlg, SW_SHOWNORMAL);
				SetForegroundWindow(hDlg);
			}

			break;
		}
		default:
			break;
		}
		break;

	// ----------------------------------------------------
	// Command notify
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
			EndDialog(hDlg, 0);
			return TRUE;
		
		case IDCANCEL:
			ShowWindow(hDlg, SW_HIDE);
			return TRUE;

		case IDC_BTN_HIDE:
			ShowWindow(hDlg, SW_HIDE);
			return TRUE;

		case IDC_BTN_SETTING:
			DialogBox(g_hInst, MAKEINTRESOURCE(IDD_DOCSETTING), hDlg, SettingDlgProc);
			return TRUE;
		}
		break;

	case WM_HELP:
			DialogBox(g_hInst, MAKEINTRESOURCE(IDD_HELP), hDlg, HelpDlgProc);
			return TRUE;

	// ----------------------------------------------------
	// Timer for shutting down animation
	case WM_TIMER:
		if( WAIT_OBJECT_0 == WaitForSingleObject(pListenThreadPack->hThread, 0) )
		{
			KillTimer( hDlg, 1 );
			EndDialog( hDlg, 1 );
		}
		else
		{

			WCHAR wszBuf[32];
			nShutdownGoing++;
			int mark = nShutdownGoing%2 ? '|': '-';
			wsprintf( wszBuf, L"Shutdown...%c", mark );
			SetDlgItemText(hDlg, IDC_MESSAGE, wszBuf);

			if( mark > 100 )
			{
				TerminateThread(pListenThreadPack->hThread, 0);
				KillTimer( hDlg, 1 );
				EndDialog( hDlg, 1 );
			}
		}
		return TRUE;

	case WM_CLOSE:
		pListenThreadPack->bValid = 0;

		if( nShutdownGoing == 0 )
		{
			nShutdownGoing = '-';
			SetTimer( hDlg, 1, 100, NULL );
		}
		return TRUE;

	// ----------------------------------------------------
	// Shutdown
	case WM_DESTROY:
		Shell_NotifyIcon(NIM_DELETE, &nIcon);
		// Wait for end of threads.
		pListenThreadPack->bValid = FALSE;
	
		if(WAIT_TIMEOUT == WaitForSingleObject(pListenThreadPack->hThread, 30000))
		{
			// Terminate threads with time up.
			TerminateThread(pListenThreadPack->hThread, 0);
		}

		CloseHandle(pListenThreadPack->hThread);
		FREEIF( pListenThreadPack );

		return 0;
	}

	return FALSE;
}

//-----------------------------------------------------------------------------
//	@detail
//		Initialize Setting dialog box.
//	@param
//	@return
//-----------------------------------------------------------------------------
static BOOL ShowSetting(HWND hDlg)
{
	WCHAR	tszValue[MAX_PATH] = {0};

	// ----------------------------------------------------------
	// show document root
	MultiByteToWideChar(CP_ACP, 0, g_conf.m_pszDocRootName, -1, tszValue, MAX_PATH );
	SetDlgItemText(hDlg, IDC_DOCROOT, tszValue );

	// ----------------------------------------------------------
	// show cgi root
	MultiByteToWideChar(CP_ACP, 0, g_conf.m_pszCgiRootName, -1, tszValue, MAX_PATH );
	SetDlgItemText(hDlg, IDC_CGIROOT, tszValue );

	// ----------------------------------------------------------
	// Listen port
	SetDlgItemInt(hDlg, IDC_LISTENPORT, g_conf.m_nListenPort, 0);

	// ----------------------------------------------------------
	// Servers
	SetDlgItemInt(hDlg, IDC_SERVERS, g_conf.m_nStartServers, 0);

	return TRUE;
}

//-----------------------------------------------------------------------------
//	@detail
//		About dialog
//	@param
//	@return
//-----------------------------------------------------------------------------
static BOOL CALLBACK HelpDlgProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	switch(Msg)
	{
	// ----------------------------------------------------------
	case WM_INITDIALOG:
	{
		WCHAR* pwszServerName = malloc_char2wchar( g_conf.m_pszServerSoftware );
		SetDlgItemText( hDlg, IDC_SERVERNAME, pwszServerName );
		FREEIF( pwszServerName );
		return TRUE;
	}

	// ----------------------------------------------------------
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
		case IDCANCEL:
			EndDialog(hDlg, 0);
			return TRUE;
		}
		break;

	// ----------------------------------------------------------
	case WM_CLOSE:
		EndDialog(hDlg, 0);
		return TRUE;
	}
	
	return FALSE;
}

//-----------------------------------------------------------------------------
//	@detail
//		Setting dialog
//	@param
//	@return
//-----------------------------------------------------------------------------
static BOOL CALLBACK SettingDlgProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	static	LISTEN_THREAD* pListenThreadPack = NULL;
	static	NOTIFYICONDATA	nIcon;
	static	HICON			hIconTaskBar;

	switch(Msg)
	{
	// ----------------------------------------------------------
	case WM_INITDIALOG:
	{
		ShowSetting(hDlg);
		return TRUE;
	}

	// ----------------------------------------------------------
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
			if( IDYES == 
				MessageBox(hDlg, L"May I save this setup?", 
					WSTR_MSGBOX_TILTE, MB_YESNO|MB_ICONWARNING|MB_DEFBUTTON2 ) )
			{
				if( !SetSetting(hDlg) )
				{
					MessageBox(hDlg, L"Save setup failed.",  
						WSTR_MSGBOX_TILTE, MB_OK|MB_ICONERROR );
					return TRUE;
				}
			}
			EndDialog(hDlg, 0);
			return TRUE;
		
		case IDCANCEL:
			EndDialog(hDlg, 0);
			return TRUE;
		}
		break;

	// ----------------------------------------------------------
	case WM_CLOSE:
		EndDialog(hDlg, 0);
		return TRUE;
	}

	return FALSE;
}

//-----------------------------------------------------------------------------
//	@detail
//		Save settings
//	@param
//	@return
//-----------------------------------------------------------------------------
static BOOL SetSetting(HWND hDlg)
{
	WCHAR	tszBuff[MAX_PATH] = {0};
	WCHAR*	pt;
	char	szDoc[MAX_PATH] = {0};
	char	szCgi[MAX_PATH] = {0};
	unsigned int unPort;
	unsigned int unServers;

	// --------------------------------------------------------------
	// document root
	// --------------------------------------------------------------
	pt = &tszBuff[0];
	GetDlgItemText(hDlg, IDC_DOCROOT, tszBuff, MAX_PATH-2);

	while( *pt != '\0' ) { pt++; }
	pt--;
	if( *pt == '\\' ) { *pt = '\0'; }
	
	WideCharToMultiByte(CP_ACP, 0, tszBuff, -1, szDoc, MAX_PATH, NULL, NULL);

	// --------------------------------------------------------------
	// extention root
	// --------------------------------------------------------------
	pt = (WCHAR*)memset( tszBuff, 0, sizeof(WCHAR)*MAX_PATH );
	GetDlgItemText(hDlg, IDC_CGIROOT, tszBuff, MAX_PATH-2);

	while( *pt != '\0' ){ pt++; }
	pt--;
	if( *pt == '\\' ){ *pt = '\0'; }
	
	WideCharToMultiByte(CP_ACP, 0, tszBuff, -1, szCgi, MAX_PATH, NULL, NULL);

	if( 0 != STRNICMP( szDoc, szCgi, strlen( szDoc ) ) )
	{
		MessageBox(hDlg, TEXT(
			"Please be sure to place a ExtRoot under DocRoot."
			), WSTR_MSGBOX_TILTE, MB_ICONHAND);
		return FALSE;
	}
	// --------------------------------------------------------------
	// listen port
	// --------------------------------------------------------------
	unPort = GetDlgItemInt(hDlg, IDC_LISTENPORT, NULL, FALSE);

	if( unPort < 1 || unPort > 0xFFFE )
	{
		MessageBox(hDlg, TEXT(
			"Listen port must be between 1 and 65534."
			), WSTR_MSGBOX_TILTE, MB_ICONHAND);
		return FALSE;
	}

	// --------------------------------------------------------------
	// Servers port
	// --------------------------------------------------------------
	unServers = GetDlgItemInt(hDlg, IDC_SERVERS, NULL, FALSE);

	if( unServers < 1 || unServers > 16 )
	{
		MessageBox(hDlg, TEXT(
			"Server threads must be between 1 and 16."
			), WSTR_MSGBOX_TILTE, MB_ICONHAND);
		return FALSE;
	}
	
	// --------------------------------------------------------------
	// save
	// --------------------------------------------------------------
	if( !(g_conf.WriteStringValue(MYCONF_WSTR_KEY_DOCROOT, szDoc)) )
		return FALSE;

	if( !(g_conf.WriteStringValue(MYCONF_WSTR_KEY_CGIROOT, szCgi)) )
		return FALSE;

	if( !(g_conf.WriteIntValue(MYCONF_WSTR_KEY_PORT, unPort)) )
		return FALSE;

	if( !(g_conf.WriteIntValue(MYCONF_WSTR_KEY_SERVERS, unServers)) )
		return FALSE;

	// --
	MakeLand(hDlg);
	MessageBox(hDlg, L"This setup becomes effective after a restart.", WSTR_MSGBOX_TILTE, MB_ICONINFORMATION);

	return TRUE;
}

//-----------------------------------------------------------------------------
//	@detail
//		Make document root if not there.
//	@param
//	@return
//-----------------------------------------------------------------------------
static BOOL MakeLand(HWND hDlg)
{
	HANDLE fff;
	WIN32_FIND_DATA wfd;
	char	szFile[MAX_PATH] = {0};
	WCHAR	tszFile[MAX_PATH] = {0};

	char*	pszIndexPage =
		"<HTML><BODY><H1>Can you see me?</H1><BR><HR>CEHTTP2 by"
		"<A HREF=\"http://www001.upp.so-net.ne.jp/ishi/\">IshiSoft</A></BODY></HTML>";

	STRCPY(szFile, g_conf.m_pszDocRootName);
	if( 0 == MultiByteToWideChar(CP_ACP,0,szFile,-1,tszFile, MAX_PATH) )
	{
		return FALSE;
	}

	// Finding file.
	fff = FindFirstFile( tszFile, &wfd );

	if( fff == INVALID_HANDLE_VALUE ||
		(fff != INVALID_HANDLE_VALUE) && !(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) )
	{
		FindClose( fff );

		if( IDNO == 
			MessageBox(hDlg, TEXT("May I create default directory?"), 
				WSTR_MSGBOX_TILTE, MB_YESNO|MB_ICONQUESTION|MB_DEFBUTTON1 ) )
		{
			return FALSE;
		}
		// ------------------------------------------------
		// Make document root.
		// ------------------------------------------------
		if( 0 == CreateDirectory( tszFile, NULL ) )
			return FALSE;
		// ------------------------------------------------
		// Make default document.
		// ------------------------------------------------
		STRCAT(szFile, "\\");
		STRCAT(szFile, g_conf.m_pszDefaultDocName);
		if( 0 == MultiByteToWideChar(CP_ACP,0,szFile,-1,tszFile, MAX_PATH) )
		{
			return FALSE;
		}

		HANDLE hFile = CreateFile(tszFile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		
		if( hFile == INVALID_HANDLE_VALUE )
		{
			return FALSE;
		}

		DWORD dwWritebytes;
		WriteFile(hFile, pszIndexPage, strlen(pszIndexPage), &dwWritebytes, NULL);
		CloseHandle(hFile);

		// ------------------------------------------------
		// Make cgi root.
		// ------------------------------------------------
		STRCPY(szFile, g_conf.m_pszCgiRootName);
		if( 0 == MultiByteToWideChar(CP_ACP,0,szFile,-1,tszFile, MAX_PATH) )
		{
			return FALSE;
		}
		
		if( 0 == CreateDirectory( tszFile, NULL ) )
		{
			return FALSE;
		}
	}
	else
	{
		FindClose(fff);
	}
	return TRUE;
}